/* @(#) sm_error.h 1.7 3/24/98 10:44:38 */
//
// sm_error.h
// This include file defines the error handling macros.
//  The use of these macros is described below: examples are shown after that
//
//      - Each routine MUST start and end with SME_SETUP and some variation
//        of SME_FINISH... in order to provide a stack dump capability
//        everywhere.
//      - It is expected that any operation that requires an explicit
//        indication of failure conditionally check for failure and
//        thow an exception using "SME_THROW(errorNumber, lpszError, pStrDat)" 
//        where "errorNumber" is the desired return code, "lpszError"
//        indicates the cause of the error, and pStrDat is a CSM_Str_struct
//        instance indicating any optional binary data to be returned.
//        Both "lpszError" and "pStrDat" may be NULL when invoking the macro.
//      - All conditional statements MUST throw an error exception on
//        failure.
//      - All single line function or member function executions MUST
//        be wrapped using the macro "SME(function(...))" in order to 
//        automatically indicate a full stack dump in case of any failure
//        in the called routine(s).  In these cases there will be no specific
//        string for the failure, but simply a line number, function name,
//        and the file name.  It is assumed that the underlying routine has
//        indicated the precise reason for the failure.
//
//  If a developer wishes to indicate more info on failure of a function
//  call, an explicit thow using "SME_THROW(errorNumber, lpszError, pStrDat)" 
//  can be executed at any time instead of wrapping the function in 
//  "SME(function(...))".
//
//  func () {
//     SME_SETUP("func");
//     SME(func1); // this is the body of the function
//     SME_FINISH_CATCH
//  }
//  long func1() {
//     SME_SETUP("func1");
//     if (error)
//        SME_THROW(22, "this is a local error", NULL);
//     SME(func2); 
//     SME_FINISH_CATCH
//     return whatever;
//  }
//  long func2() {
//     SME_SETUP("func2");
//     if (error)
//        SME_THROW(22, "this is the error", NULL);
//     SME_FINISH
//     SME_CATCH_SETUP
//        // cleanup logic
//     SME_CATCH_FINISH
//     return whatever;
//  }

#include "sm_internal.h"

//////////////////////////////////////////////////////////////////////////
// General error information class
class CSM_ErrorBuf {
public:
    long m_lErrorCode;     // Identifies the error and how to interpret
                           //      * the "errorBuf".
    CSM_Buffer m_ErrorBuf; // May be an ASN.1 encoded buffer.
    char *m_pszDebug;      // Debug string; cummulative.
};

//////////////////////////////////////////////////////////////////////////
// The SME_SETUP variables are used by the exception handler for stack dump
// information.  The macro minimizes the developer definitions.
// pszFuncName should be passed as a null terminated string containing
// the name of the function or member
#define SME_SETUP(pszFuncName) long lSMELine=-1; bool bSMELocal=false; \
    char *SME_FN=pszFuncName; try   {

//////////////////////////////////////////////////////////////////////////
// The SME_FINISH_CATCH macro is intended for those simple routines that 
//  require no cleanup operations.  If an exception occurs, this macro will
//  simply close the "try{}" block, invoke the "catch()" operation and 
//  finally invoke the next level exception with the received "CSM_Exception"
//  instance indicating the original cause of the error.  
#define SME_FINISH_CATCH        } SME_CATCH_SETUP SME_CATCH_FINISH

//////////////////////////////////////////////////////////////////////////
// SME_FINISH_CATCH_APP macro is intended for routines that require
// no cleanup and want to report the stack to stdout.
#define SME_FINISH_CATCH_APP    } SME_CATCH_SETUP SME_CATCH_FINISH_APP

//////////////////////////////////////////////////////////////////////////
//  These macros handle the case where a developer wishes to perform cleanup 
//  on an exception before invoking the next level exception.  Here, each 
//  operation must be explicitely started and ended.  NOTICE that these macros
//  are used by the above SME_FINISH_CATCH simple case macro.

//////////////////////////////////////////////////////////////////////////
// this macro ends the SME_SETUP block which actually ends the try block
#define SME_FINISH              }

//////////////////////////////////////////////////////////////////////////
// this macro starts the catch block (catches a CSM_Exception)
#define SME_CATCH_SETUP         catch (CSM_Exception *pException) {

//////////////////////////////////////////////////////////////////////////
// this macro completes the catch block.  If SME_SETUP was used, then
// it moves the CSM_Exception thrown information into the CSM_ErrorBuf.
// Otherwise, it adds the stack information to CSM_Exception.
// Then this macro throws CSM_Exception again...
#define SME_CATCH_FINISH        \
      if (bSMELocal)\
        pException->FillErrorBuf(NULL);\
      else\
        pException->AppendStackError(lSMELine, SME_FN, __FILE__);\
      throw (pException);\
   }

//////////////////////////////////////////////////////////////////////////
// This macro finishes the catch block, deletes CSM_Exception, and prints
// the stack to standard out.
#define SME_CATCH_FINISH_APP        \
   if (bSMELocal)\
      pException->FillErrorBuf(NULL);\
   else\
      pException->AppendStackError(lSMELine, SME_FN, __FILE__);\
   printf("ERROR INFORMATION: %s\n", pException->m_szErrorString);\
   printf("\tERROR CODE: %d\tSOURCE LINE NUMBER: %d\n",\
         pException->m_Error_code, pException->m_lLine);\
   printf("\tFILE: %s\n", pException->m_szFileName);\
   printf("\tFUNCTION: %s\n", pException->m_szFunctionName);\
   printf("ERROR STACK:\n%s\n", pException->m_szStackString);\
   delete pException;\
}

//////////////////////////////////////////////////////////////////////////
// This macro sets the line number and calls the provided function
#define SME(function) { lSMELine = __LINE__; function; }

//////////////////////////////////////////////////////////////////////////
// On the SME_THROW macro, both the pszError and pStrDat can be NULL, in-line.
// The "bSMELocal" flag indicates that this exception was thrown in the local 
// routine; this allows the same "catch()" handler to handle all exceptions.
#define SME_THROW(errorNumber, pszError, pStrDat) {\
   bSMELocal = true; throw (new CSM_Exception(errorNumber,__LINE__,SME_FN,\
      __FILE__,pszError,pStrDat)); }

//////////////////////////////////////////////////////////////////////////
// CSM_Exception is the class used to throw exceptions
class CSM_Exception {
public:
   SM_RET_VAL m_Error_code;
   long m_lLine;
   char m_szFunctionName[80];
   char m_szFileName[256];
   char m_szErrorString[1024];
   char m_szStackString[1024];
   CSM_Buffer *m_pData;

   // append error information to m_szStackString
   void AppendStackError(long lLine, char *pszFunction, char *pszFile) 
   {
      char *pszTmp;
      int nLen;
      if ((nLen=strlen(m_szStackString)))
         pszTmp = &m_szStackString[nLen];
      else
         pszTmp = &m_szStackString[0];
      if (nLen+50 < 1024)
         sprintf(pszTmp, "%s IN %s, line %ld.\n", pszFunction, 
               pszFile, lLine);
   }

   // move information from here to the provided CSM_ErrorBuf
   void FillErrorBuf(CSM_ErrorBuf *pError) 
   {
      char szDebug[256];
      char *pszSave;
      if (pError != NULL)
      {
         // put error code in
         pError->m_lErrorCode = m_Error_code;
         // put data in
         pError->m_ErrorBuf = *m_pData;
         // format debug data
         // TBD, may want to add time stamp and any other stuff
         sprintf(szDebug, "%s - %s - %ld", m_szFunctionName,
               m_szFileName, m_lLine);
         if (pError->m_pszDebug == NULL)
            pError->m_pszDebug = strdup(szDebug);
         else
         {
            pszSave = pError->m_pszDebug;
            pError->m_pszDebug = (char *)calloc(1, strlen(pszSave) +
                     strlen(szDebug) + 4);
            sprintf(pError->m_pszDebug, "%s\n%s", szDebug, pszSave);
            free(pszSave);
         }
      }   
   }

   CSM_Exception(SM_RET_VAL err_num, long l, char *pszFuncName,
      char *pszFileName, char *pszErrorString, CSM_Buffer *d) 
   {
         m_Error_code = err_num;
         m_lLine = l;
         strcpy(m_szFunctionName, pszFuncName);
         strcpy(m_szFileName, pszFileName);
         if (pszErrorString != NULL)
            strcpy(m_szErrorString, pszErrorString);
         else
            m_szErrorString[0] = '\0';
         if (d)
            m_pData = new CSM_Buffer(*d);
         else
            m_pData = NULL;
         m_szStackString[0] = '\0';
   }

   ~CSM_Exception() 
   {
       if (m_pData)
         delete m_pData;
   }
};

// EOF sm_error.h
