/* @(#) sm_free.h 1.16 3/24/98 10:43:56 */

//////////////////////////////////////////////////////////////////////////
// sm_free.h
// 
// header file implementing freely distributable S/MIME v3 MUST algorithms
// to include SHA1, DSA, DH, and 3DES.
//
// Should be directly included by the C++ app and the sm_free CTI Library
//
//////////////////////////////////////////////////////////////////////////

#ifndef _SM_FREE_H_
#define _SM_FREE_H_

// constant defines used in the free CTI
#define SM_FREE_BLOCK_SIZE 32767
#define SM_FREE_3DES_KEYLEN 24
#define SM_FREE_3DES_IVLEN 8
#define SM_FREE_3DES_BLOCKSIZE 8
#define SM_FREE_RA_SIZE 128
#define SM_FREE_DSA_SIG_LEN 40
#define SM_FREE_FORTENC "FORTEZZA MEK: REG"

//////////////////////////////////////////////////////////////////////////
// SM_FREE errors 6000-6999

// parameter missing to a SMTI function
#define SM_FREE_MISSING_PARAM       6000
// occurs when unable to ASN.1 decode provided parameters
#define SM_FREE_PARAM_DEC_ERROR     6001
// occurs when caller requests CSM_Free to process or operate using an
// unsupported algorithm
#define SM_FREE_UNSUPPORTED_ALG     6002
// occurs when SMTI_Verify signature check fails
#define SM_FREE_VERIFY_FAILED       6003
// occurs when CSM_Free doesn't have a preferred digest alg set
#define SM_FREE_NO_DIGEST_ALG       6004
// occurs when CSM_Free can't find issuer of a user cert in the address book
#define SM_FREE_MAB_NO_ISSUER       6005
// occurs when the issuer from the address book is not a DSA cert
#define SM_FREE_ISSUER_NOT_DSA      6006
// occurs when the attempt to load a public value into the CSM_Free fails
#define SM_FREE_PUT_Y_ERROR         6007
// occurs when the pad values in the final block of encrypted data are
// incorrect...would probably occur if decrypt used the wrong MEK
#define SM_FREE_DECRYPT_PAD_ERROR   6008

// SM_FREE Pad macro
#define SM_FREE_Pad(buf, size) {\
      unsigned int b = SM_FREE_3DES_BLOCKSIZE - \
            (size % SM_FREE_3DES_BLOCKSIZE); \
      memset(buf+size, (char) b, (size_t) b); \
      size += b;                     \
      }

#include "sm_api.h"
#include "sm_common.h"

#include "rng.h" // random number generator
#include "des.h" // TripleDES
#include "modes.h" // CBCEncryption
#include "integer.h" // big integers
#include "dh.h" // DH key agreement
#include "dsa.h" // DSA digest encryption
#include "sha.h" // SHA1
#include "md5.h" // MD5

class CSM_Free : public CSM_Common
{
public:
   SM_RET_VAL SMTI_Login(void);
   SM_RET_VAL SMTI_Sign(
            CSM_Buffer *pData, // input, data to be signed
            CSM_Buffer *pEncryptedDigest, // signature
            CSM_Buffer *pDigest); // digest
   SM_RET_VAL SMTI_Verify(
            CSM_Buffer *pSignerKey, // input
            CSM_Alg    *pDigestAlg, // input
            CSM_Alg    *pSignatureAlg, // input
            CSM_Buffer *pData, // input
            CSM_Buffer *pSignature); // input
   SM_RET_VAL SMTI_Encrypt(
            CSM_Buffer *pData, // input (data to be encrypted)
            CSM_Buffer *pEncryptedData, // output
            CSM_Buffer *pParameters, // output
            CSM_Buffer *pMEK); // output
   SM_RET_VAL SMTI_GenerateEMEK(
            CSM_Buffer *pRecipient, // input, Y of recip
            CSM_Buffer *pParameters, // input
            CSM_Buffer *pMEK, // input, MEK or special phrase
            CSM_Buffer *pEMEK, // output, encrypted MEK
            CSM_Buffer *pUKM); // output
   SM_RET_VAL SMTI_ExtractMEK(
            CSM_Buffer *pOriginator, // input, Y of originator
            CSM_Buffer *pParameters, // input
            CSM_Buffer *pEMEK, // input, encrypted MEK
            CSM_Buffer *pUKM, // input
            CSM_Buffer *pMEK); // output
   SM_RET_VAL SMTI_Decrypt(
            CSM_Buffer *pParameters, // input (initialization vector)
            CSM_Buffer *pEncryptedData, // input (data to be decrypted)
            CSM_Buffer *pMEK, // input (MEK or special phrase)
            CSM_Buffer *pData); // output (decrypted data)
   SM_RET_VAL SMTI_DigestData(
            CSM_Buffer *pData, // input
            CSM_Buffer *pDigest); // output
   SM_RET_VAL SMTI_Random(
            CSM_Buffer *pSeed, // input
            CSM_Buffer *pRandom, // input/output
            SM_SIZE_T lLength); // input
   // this CTI does nothing with locking/unlocking 
   SM_RET_VAL SMTI_Lock() { return SM_NO_ERROR; };
   SM_RET_VAL SMTI_Unlock() { return SM_NO_ERROR; };

   // This CTI will override BTIFindAlgIds because in CEA and KEA's
   // case, we only want to compare OIDs and are not interested in 
   // comparing the parameters part of the AlgId
   bool BTIFindAlgIds(CSM_Alg *pdigestAlgID, 
            CSM_Alg *pdigestEncryptionAlgID,
            CSM_Alg *pkeyEncryptionAlgID,
            CSM_Alg *pcontentEncryptionAlgID);

   CSM_Free(); // default constructor
   ~CSM_Free(); // DESTRUCTOR
   void FreeAB(); // free AB no matter what
   void FreeAB(CSMIME *pCSMIME); // free AB if its the only one

   // set or restore the default Alg OIDs
   void SetDefaultOIDs();

   // member access functions
   char *GetPassword();
   void SetPassword(char *pszPassword);
   MAB_AB_def *GetAB();
   void SetAB(MAB_AB_def *pAB);

   // called by init
   SM_RET_VAL GetParamsAndY(MAB_Entrydef *pEntry, MAB_AB_def *pAB,
         AlgorithmIdentifier *&pAlgID);

   // TBD, following should be private, used for testing now...
   CSM_Buffer m_X;         // DH Private Key
   CSM_Buffer m_ParamP;    // DH P Parameter
   CSM_Buffer m_ParamG;    // DH G Parameter
   Integer m_DSAP;      // DSA P Parameter
   Integer m_DSAQ;      // DSA Q Parameter
   Integer m_DSAG;      // DSA G Parameter
   Integer m_DSAY;      // DSA Public Key
   Integer m_DSAX;      // DSA Private Key

   X917RNG *m_pRng; // random number generator
   char *m_pszPreffix; // ID Preffix specified by app

private:
   char *m_pszPassword;
   MAB_AB_def *m_pAB;

   BlockTransformation *m_pRandomCipher;
   CSM_Buffer m_seed;

   // Extract X returns a ptr to the current private key
   // and sets the length in lXLen
   unsigned char * ExtractX(long &lXLen);

   SM_RET_VAL ExtractParams(AlgorithmIdentifier *pAlgID);
   SM_RET_VAL DecodeCertificate(CSM_Buffer *pEncodedCert,
         Certificate *pSnaccCertificate, Name **ppIssuer,
         AlgorithmIdentifier **ppAlgID, AsnBits **ppY);
};

SM_RET_VAL SMFreeInit(CSMIME *pCSMIME, char *pszPassword,
                     char *pszAddressBook, char *pszPreffix);

#endif // _SM_FREE_H_

// EOF sm_free.h
