/* @(#) sm_rsa.h 1.5 3/25/98 13:40:36 */
//////////////////////////////////////////////////////////////////////////
// sm_rsa.h
// 
// header file implementing BSAFE algorithms to include RSA, RC2, MD5 and SHA1.
//
// Should be directly included by the C++ app and the sm_rsa CTI Library
//
//////////////////////////////////////////////////////////////////////////

#ifndef _SM_RSA_H_
#define _SM_RSA_H_

// constant defines used in the rsa CTI
#define SM_RSA_BLOCK_SIZE 32767

//////////////////////////////////////////////////////////////////////////
// SM_RSA errors 5000-5999

// parameter missing to a SMTI function
#define SM_RSA_MISSING_PARAM       5000
// occurs when unable to ASN.1 decode provided parameters
#define SM_RSA_PARAM_DEC_ERROR     5001
// occurs when caller requests CSM_Rsa to process or operate using an
// unsupported algorithm
#define SM_RSA_UNSUPPORTED_ALG     5002
// occurs when SMTI_Verify signature check fails
#define SM_RSA_VERIFY_FAILED       5003
// occurs when CSM_Rsa doesn't have a preferred digest alg set
#define SM_RSA_NO_DIGEST_ALG       5004
// occurs when CSM_Rsa can't find issuer of a user cert in the address book
#define SM_RSA_MAB_NO_ISSUER       5005
// occurs when the issuer from the address book is not a DSA cert
#define SM_RSA_ISSUER_NOT_DSA      5006
// occurs when the attempt to load a public value into the CSM_Rsa fails
#define SM_RSA_PUT_Y_ERROR         5007

// SM_RSA Pad macro
#define SM_RSA_Pad(buf, size) {\
      unsigned int b = SM_RSA_3DES_BLOCKSIZE - \
            (size % SM_RSA_3DES_BLOCKSIZE); \
      memset(buf+size, (char) b, (size_t) b); \
      size += b;                     \
      }

#include "sm_api.h"
#include "aglobal.h"
#include "bsafe.h"


class CSM_Rsa : public CSM_BaseTokenInterface
{
public:
   SM_RET_VAL SMTI_Login(void);
   SM_RET_VAL SMTI_Sign(
            CSM_Buffer *pData, // input, data to be signed
            CSM_Buffer *pEncryptedDigest, // signature
            CSM_Buffer *pDigest); // digest
   SM_RET_VAL SMTI_Verify(
            CSM_Buffer *pSignerKey, // input
            CSM_Alg    *pDigestAlg, // input
            CSM_Alg    *pSignatureAlg, // input
            CSM_Buffer *pData, // input
            CSM_Buffer *pSignature); // input
   SM_RET_VAL SMTI_Encrypt(
            CSM_Buffer *pData, // input (data to be encrypted)
            CSM_Buffer *pEncryptedData, // output
            CSM_Buffer *pParameters, // output
            CSM_Buffer *pMEK); // output
   SM_RET_VAL SMTI_GenerateEMEK(
            CSM_Buffer *pRecipient, // input, Y of recip
            CSM_Buffer *pParameters, // input
            CSM_Buffer *pMEK, // input, MEK or special phrase
            CSM_Buffer *pEMEK, // output, encrypted MEK
            CSM_Buffer *pUKM); // output
   SM_RET_VAL SMTI_ExtractMEK(
            CSM_Buffer *pOriginator, // input, Y of originator
            CSM_Buffer *pParameters, // input
            CSM_Buffer *pEMEK, // input, encrypted MEK
            CSM_Buffer *pUKM, // input
            CSM_Buffer *pMEK); // output
   SM_RET_VAL SMTI_Decrypt(
            CSM_Buffer *pParameters, // input (initialization vector)
            CSM_Buffer *pEncryptedData, // input (data to be decrypted)
            CSM_Buffer *pMEK, // input (MEK or special phrase)
            CSM_Buffer *pData); // output (decrypted data)
   SM_RET_VAL SMTI_DigestData(
            CSM_Buffer *pData, // input
            CSM_Buffer *pDigest); // output
   SM_RET_VAL SMTI_Random(
            CSM_Buffer *pSeed, // input
            CSM_Buffer *pRandom, // input/output
            SM_SIZE_T lLength); // input
   // this CTI does nothing with locking/unlocking 
   SM_RET_VAL SMTI_Lock() { return SM_NO_ERROR; };
   SM_RET_VAL SMTI_Unlock() { return SM_NO_ERROR; };

   CSM_Rsa(); // default constructor
   ~CSM_Rsa(); // DESTRUCTOR
   void FreeAB(); // free AB no matter what
   void FreeAB(CSMIME *pCSMIME); // free AB if its the only one

   // set or restore the default Alg OIDs
   void SetDefaultOIDs();

   // member access functions
   char *GetPassword();
   void SetPassword(char *pszPassword);
   MAB_AB_def *GetAB();
   void SetAB(MAB_AB_def *pAB);

   // called by init
   SM_RET_VAL GetParamsAndY(MAB_Entrydef *pEntry, MAB_AB_def *pAB,
      AlgorithmIdentifier *&pAlgID);

   // random number generator
   B_ALGORITHM_OBJ randomAlgorithm;

private:
   char *m_pszPassword;
   MAB_AB_def *m_pAB;
   CSM_Buffer m_seed;
};

#endif // _SM_RSA_H_

// EOF sm_rsa.h
