#ifndef NO_SCCS_ID
static char SccsId[ ] = "@(#) sm_Encrypt.cpp 1.14 3/25/98 17:28:27"; 
#endif

//////////////////////////////////////////////////////////////////////////
// sm_Encrypt.cpp
// Implementation of the CSM_EncryptMsg and CSM_EncryptData classes
// CSM_EncryptMsg is for high level use.  The application developer
// should not have to directly access the snacc generated classes.
// CSM_EncryptData is for low level use.  The application may have
// to directly access the exposed snacc generated class.  Both
// classes have the purpose of generating valid CSM EnvelopedData
// ASN.1 encoded structures based on the provided input and the
// provided (and prepared) instances (contained in CSMIME)
//////////////////////////////////////////////////////////////////////////

#include "sm_api.h"

//////////////////////////////////////////////////////////////////////////
// CONSTRUCTORS
//////////////////////////////////////////////////////////////////////////
// empty constructor
CSM_EncryptMsg::CSM_EncryptMsg()
{
   Clear();
}

//////////////////////////////////////////////////////////////////////////
// incoming message is in pBlob
CSM_EncryptMsg::CSM_EncryptMsg(CSM_Buffer *pBlob)
{
   Clear();
   SetEncapContent(pBlob, CSM_Content::CONTENT_DATA_DATA);
}

//////////////////////////////////////////////////////////////////////////
// incoming message is in pContent
CSM_EncryptMsg::CSM_EncryptMsg(CSM_Content *pContent)
{
   Clear();
   SetEncapContent(*pContent);
}

//////////////////////////////////////////////////////////////////////////
// Destructor
CSM_EncryptMsg::~CSM_EncryptMsg()
{
   if (m_poidContentEncrypt)
      delete (m_poidContentEncrypt);
   if (m_pMsgCrtCrls)
      delete (m_pMsgCrtCrls);
   if (m_pRecipients)
      delete (m_pRecipients);
}

//////////////////////////////////////////////////////////////////////////
// CSM_EncryptMsg::Encrypt uses the content, m_pMsgCrtCrls
// and m_pRecipients to generate the EnvelopedData
void CSM_EncryptMsg::Encrypt(CSMIME *pCSMIME)
{
   SME_SETUP("CSM_EncryptMsg::Encrypt");
   CSM_CSInstLst *pInstTmp;
   CSM_CertificateChoiceLst *pCertificates;
   CSM_Buffer *pbufferResult;

   // check to make sure everything is ready to roll
   if ((pCSMIME == NULL) || (pCSMIME->m_pCSInsts == NULL))
      SME_THROW(SM_MISSING_PARAM, NULL, NULL);
   if (m_poidContentEncrypt == NULL)
      SME_THROW(SM_ENCRYPTION_UNPREPARED, 
            "no content encryption oid specified", NULL);
   if (m_pRecipients == NULL)
      SME_THROW(SM_ENCRYPTION_UNPREPARED, 
            "no recipients specified", NULL);

   // in order to call CSM_EncryptData::Encrypt we need pCSMIME which the
   // application gave us

   // we also need a CSM_MsgCertCrls which the application prepared.
   // if the app set m_bIncludeOrigCertsFlag then we need to put the certs
   // from each UseThis instance into the m_pMsgCrtCrls
   if (m_bIncludeOrigCertsFlag)
   {
      // the application has requested that we include the originator
      // certs, therefore, go through the instances and find all of them
      // that are marked with UseThis and include whatever certs are
      // present in the instance in this m_pMsgCrtCrls
      pInstTmp = pCSMIME->m_pCSInsts->First();
      do
      {
         // if this instance is used
         if (pInstTmp->Get()->IsThisUsed())
         {
            // if there are certificates in this instance
            if ((pCertificates = pInstTmp->Get()->GetCertificates()) != NULL)
            {
               // if this class already has certificates
               if (m_pMsgCrtCrls)
               {
                  m_pMsgCrtCrls->SetCertificates(pCertificates);
               }
               else // this class doesn't have any certificates yet
               {
                  if ((m_pMsgCrtCrls = new CSM_MsgCertCrls(pCertificates))
                           == NULL)
                     SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
               }
            }
         }
      } while ((pInstTmp = pInstTmp->Next()) != NULL);
   }

   // we will need to pass m_pRecipients to CSM_EncryptData::Encrypt, the
   //    app must have prepared this...
   // we must pass the poidContentType to CSM_EncryptData::Encrypt which
   //    is obtained via the inherited CSM_CommonData->GetEncapContent()->
   //    m_ContentType
   // we must pass the content to CSM_EncryptData::Encrypt which
   //    is obtained via the inherited CSM_CommonData->GetEncapContent()->
   //    m_content
   // we must pass the content encryption oid to CSM_EncryptData::Encrypt
   //    and the app must have set this with the SetContentEncryptOID member
   CSM_Alg *pContentEncryptionAlg = new CSM_Alg(*m_poidContentEncrypt);

   // setup a buffer to receive the result
   // TBD, what type of buffer?  for now, do memory...
   if ((pbufferResult = new CSM_Buffer) == NULL)
      SME_THROW(SM_MEMORY_ERROR, NULL, NULL);

   // everything should be ready to call CSM_EncryptData::Encrypt now...
   SME(CSM_EncryptData::Encrypt(pCSMIME, m_pMsgCrtCrls, m_pRecipients,
            &(GetEncapContent()->m_contentType), 
            &(GetEncapContent()->m_content), pContentEncryptionAlg,
            pbufferResult));

   // store the encoded blob
   SetEncodedBlob(pbufferResult);

   SME_FINISH
   SME_CATCH_SETUP
      // catch/cleanup logic as necessary
   SME_CATCH_FINISH
}

//////////////////////////////////////////////////////////////////////////
void CSM_EncryptData::AddUKMToOriginatorInfo(CSM_Buffer *pUKM,
      CSM_Alg *pAlg)
{
   OriginatorInfo *pOI;
   UserKeyingMaterial *pSnaccUKM = NULL;
   bool bFoundUKM = false;

   SME_SETUP("CSM_EncryptData::AddUKMToOriginatorInfo");

   if ((pUKM != NULL) && (pUKM->AccessAll() != NULL))
   {
      // only add a ukm if one exists and if pAlg is passed in
      if (pAlg == NULL)
         SME_THROW(SM_MEMORY_ERROR, "no alg id for ukm", NULL);

      // get a pointer to the originator info, create if necessary
      if (m_SnaccEnvelopedData.originatorInfo == NULL)
      {
         if ((pOI = m_SnaccEnvelopedData.originatorInfo = new OriginatorInfo)
               == NULL)
            SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
      }
      else
         pOI = m_SnaccEnvelopedData.originatorInfo;

      // create a new UserKeyingMaterials if it doesn't exist yet...
      if (pOI->ukms == NULL)
      {
         if ((pOI->ukms = new UserKeyingMaterials) == NULL)
            SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
      }
      else
      {
         UserKeyingMaterial *pTmpUkm;
         // look through the existing ukms to see if one for this alg id
         // already exists
         pOI->ukms->SetCurrToFirst();
         for (pTmpUkm = pOI->ukms->First(); pTmpUkm && 
               (*pAlg != pTmpUkm->algorithm->algorithm);
               pTmpUkm = pOI->ukms->GoNext());
         if (pTmpUkm)
            bFoundUKM = true;
      }

      // if we didn't find a matching ukm, add a new node and fill it
      if (!bFoundUKM)
      {
         if ((pSnaccUKM = pOI->ukms->Append()) == NULL)
            SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
         // put the ukm data into the new node
         pSnaccUKM->ukm.Set(pUKM->AccessAll(), pUKM->Length());
         // put the algorithm ID into the new node
         if ((pSnaccUKM->algorithm = new AlgorithmIdentifier) == NULL)
            SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
         pSnaccUKM->algorithm->algorithm = *pAlg->GetId()->GetSNACCOid();
      }
      // if it was found, then we don't need to add it again...
   }
      
   SME_FINISH
   SME_CATCH_SETUP
      // catch/cleanup logic as necessary
   SME_CATCH_FINISH
}

//////////////////////////////////////////////////////////////////////////
CSM_Buffer* CSM_EncryptData::GetUKMFromOriginatorInfo(CSM_Alg *pAlg)
{
   CSM_Buffer *pUKM = NULL;
   SME_SETUP("CSM_EncryptData::GetUKMFromOriginatorInfo");

   if (pAlg == NULL)
      SME_THROW(SM_MISSING_PARAM, NULL, NULL);

   if ((pUKM = new CSM_Buffer) == NULL)
      SME_THROW(SM_MEMORY_ERROR, NULL, NULL);

   if ((m_SnaccEnvelopedData.originatorInfo != NULL) &&
         (m_SnaccEnvelopedData.originatorInfo->ukms != NULL))
   {
      // loop through the existing ukms to see if one for this alg id
      // already exists
      UserKeyingMaterial *pTmpUkm;
      m_SnaccEnvelopedData.originatorInfo->ukms->SetCurrToFirst();
      for (pTmpUkm = m_SnaccEnvelopedData.originatorInfo->ukms->First();
            pTmpUkm && (*pAlg != pTmpUkm->algorithm->algorithm);
            pTmpUkm = m_SnaccEnvelopedData.originatorInfo->ukms->GoNext());
      // if it was found, store it in the return buffer
      if (pTmpUkm)
         SME(pUKM->Set((char*)pTmpUkm->ukm, pTmpUkm->ukm.Len()));
   }

   SME_FINISH_CATCH
   return pUKM;
}

//////////////////////////////////////////////////////////////////////////
// LoadFromMsgCertCrls takes the certs, ACs and CRLs from the provided
// CSM_MsgCertCrls and puts it in decoded form into the EnvelopedData
void CSM_EncryptData::LoadFromMsgCertCrls(CSM_MsgCertCrls *pMsgCertCrls)
{
   CSM_CertificateChoiceLst *pCerts;
   CSM_BufferLst *pCRLs;
   CertificateChoices *pSnaccCertChoice;
   AttributeCertificate snaccAC;
   CertificateList *pSnaccCRL;
   OriginatorInfo *pOI;
   bool bSomethingLoaded = false;

   SME_SETUP("CSM_EncryptData::LoadFromMsgCertCrls");

   if (pMsgCertCrls != NULL)
   {
      if ((pOI = m_SnaccEnvelopedData.originatorInfo = new OriginatorInfo)
            == NULL)
         SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
      // navigate pMsgCertCrls and store its components in the
      // appropriate places
      pCerts = pMsgCertCrls->GetCertificates();
      // load certificates
      if (pCerts != NULL)
      {
         // create the certificate set
         if ((pOI->certs = new CertificateSet) == NULL)
            SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
         // extract the Certs from the pCerts CSM_CertificateChoiceLst,
         // decode them, and store them in the certificateSet
         do
         {
            if ((pSnaccCertChoice = pOI->certs->Append()) == NULL)
               SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
            // set the type
            pSnaccCertChoice->choiceId = pSnaccCertChoice->certificateCid;
            // decode the current certificate
            if ((pSnaccCertChoice->certificate = new Certificate) == NULL)
               SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
            DECODE_BUF(pSnaccCertChoice->certificate,
                  (*(((pCerts->Get())->GetEncodedCert()))));
            bSomethingLoaded = true;
         } while ((pCerts = pCerts->Next()) != NULL);
      }
      pCerts = pMsgCertCrls->GetACs();
      // load ACs
      if (pCerts != NULL)
      {
         // create the certificate set if it isn't already created
         if (pOI->certs == NULL)
            if ((pOI->certs = new CertificateSet) == NULL)
               SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
         // extract the ACs from the pCerts CSM_CertificateChoiceLst,
         // decode them, and store them in the certificateSet
         do
         {
            if ((pSnaccCertChoice = pOI->certs->Append()) == NULL)
               SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
            // set the type
            pSnaccCertChoice->choiceId = pSnaccCertChoice->attrCertCid;
            // decode the current certificate
            DECODE_BUF((&snaccAC), (*(((pCerts->Get())->GetEncodedCert()))));
            if ((pSnaccCertChoice->attrCert = 
                  new AttributeCertificate(snaccAC)) == NULL)
               SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
            bSomethingLoaded = true;
         } while ((pCerts = pCerts->Next()) != NULL);
      }
      pCRLs = pMsgCertCrls->GetCRLs();
      // load CRLs
      if (pCRLs != NULL)
      {
         // create the crl list
         if ((pOI->crls = new CertificateRevocationLists) == NULL)
            SME_THROW(SM_MEMORY_ERROR, NULL, NULL);

         // move the CRLs from the pCRLs CSM_BufferLst and put them into
         // the provided certificateRevocationLists
         do
         {
            if ((pSnaccCRL = pOI->crls->Append()) == NULL)
               SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
            // decode the current CRL
            DECODE_BUF(pSnaccCRL, (*(pCRLs->Get())) );
            bSomethingLoaded = true;
         } while ((pCRLs = pCRLs->Next()) != NULL);
      }
      // TBD, do ukms

      // if nothing was loaded, clear out the variable and return 
      // without error...this means something funky was passed in
      // but oh well...
      if (!bSomethingLoaded)
      {
         delete m_SnaccEnvelopedData.originatorInfo;
         m_SnaccEnvelopedData.originatorInfo = NULL;
      }
   }

   SME_FINISH
   SME_CATCH_SETUP
      // catch/cleanup logic as necessary
      // cleanup based on bSomethingLoaded
   SME_CATCH_FINISH
}

//////////////////////////////////////////////////////////////////////////
// CSM_EncryptData::Encrypt uses the provided parameters and whatever
// is already available in the content and exposed m_SnaccEnvelopedData
// to generate the EnvelopedData
void CSM_EncryptData::Encrypt(CSMIME *pCSMIME,
                                    CSM_MsgCertCrls *pMsgCertCrls,
                                    CSM_RecipientLst *pRecipients,
                                    CSM_OID *poidContentType,
                                    CSM_Buffer *pContent,
                                    CSM_Alg *pContentEncryptionAlg,
                                    CSM_Buffer *pOutputBuf)
{
   SM_RET_VAL lRet = SM_NO_ERROR;
   CSM_Buffer bufContentParameters, bufEncryptedContent, bufMEK;
   CSM_Buffer bufParams, bufEMEK;
   CSM_Buffer *pbufUKM;
   long lVersion = SM_ENV_DATA_VERSION;
   RecipientInfo *pRecipientInfo = NULL;
   CSM_CSInst *pTmpInstance;
   CSM_RecipientLst *pTmpRecipLst;
   CSM_CSInstLst *pTmpInstLst;
   EncryptedContentInfo *pECI; // temporary snacc pointer
   CSM_Buffer *pbufSerial;

   SME_SETUP("CSM_EncryptData::Encrypt");

   // check incoming parameters and values
   if ((pCSMIME == NULL) || (pContent == NULL) || (pOutputBuf == NULL) ||
         (pRecipients == NULL) || (pContentEncryptionAlg == NULL) ||
         (pCSMIME->m_pCSInsts == NULL))
      SME_THROW(SM_MISSING_PARAM, NULL, NULL);
   if (pContent->AccessAll() == NULL)
      SME_THROW(SM_ENCRYPTION_UNPREPARED,
            "provided content is empty", NULL);
   if (pContentEncryptionAlg->m_pId == NULL)
      SME_THROW(SM_ENCRYPTION_UNPREPARED,
            "provided content encryption alg empty", NULL);

   // find the first pCSMIME->instance that can do the requested
   // content encryption algorithm
   if ((pTmpInstance = pCSMIME->FindCSInstAlgIds(NULL, NULL, NULL,
            pContentEncryptionAlg)) == NULL)
      SME_THROW(SM_NO_SUPPORTING_INSTANCE, 
            "no instance supports requested cont encr alg", NULL);

   // lock the CTI
   SME(pCSMIME->InstanceLock(SM_INST_USE_THIS));

   ///////////////////////////////
   // encrypt the provided content
   SME(pTmpInstance->GetTokenInterface()->SMTI_Encrypt(pContent,
         &bufEncryptedContent, &bufContentParameters, &bufMEK));

   ///////////////////////////////////////////////////////
   // load the encrypted content into m_SnaccEnvelopedData
   if ((pECI = m_SnaccEnvelopedData.encryptedContentInfo = 
         new EncryptedContentInfo) == NULL)
      SME_THROW(SM_MEMORY_ERROR, NULL, NULL);

   // store content type
   pECI->contentType.Set(*(poidContentType->GetSNACCOid()));
   // store alg OID
   if ((pECI->contentEncryptionAlgorithm = 
         new ContentEncryptionAlgorithmIdentifier) == NULL)
      SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
   pECI->contentEncryptionAlgorithm->algorithm.Set
         (*(pContentEncryptionAlg->GetId()->GetSNACCOid()));
   // store alg parameters
   if (bufContentParameters.AccessAll() != NULL)
   {
      if ((pECI->contentEncryptionAlgorithm->parameters = new AsnAny) == NULL)
         SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
      SM_ASSIGN_ANYBUF((&bufContentParameters), 
            pECI->contentEncryptionAlgorithm->parameters);
   }
   // store encrypted content
   if ((pECI->encryptedContent = new EncryptedContent) == NULL)
      SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
   pECI->encryptedContent->Set(bufEncryptedContent.AccessAll(), 
         bufEncryptedContent.Length());
   
   ///////////////////////////////////////////////
   // load up OriginatorInfo based on pMsgCertCrls
   // that were provided by the caller
   SME(LoadFromMsgCertCrls(pMsgCertCrls));

   /////////////////////////////////////////////////
   // go through pRecipients and load RecipientInfos
   pTmpRecipLst = pRecipients->First();
   do
   {
      // setup recipient's Alg
      CSM_Alg algTmp(*(pTmpRecipLst->Get()->GetPublicKeyAlg()));

      // find a UseThis instance that support's this recipient's KM Alg
      pTmpInstLst = pCSMIME->m_pCSInsts->First();
      do
      {
         // is this instance marked for use?
         if (!(pTmpInstLst->Get()->IsThisUsed()))
            continue; // no, go to next instance

         // is this recipient's alg supported by this instance?
         if ((pTmpInstLst->Get())->FindAlgIds(NULL, NULL,
               &algTmp, NULL))
         {
            // get the UKM
            pbufUKM = GetUKMFromOriginatorInfo(&algTmp);
            // found UseThis instance that supports the recipient's
            // alg, generate the EMEK
            SME((pTmpInstLst->Get()->GetTokenInterface())->
                  SMTI_GenerateEMEK(pTmpRecipLst->Get()->GetPublicKey(), 
                        &bufParams, &bufMEK, &bufEMEK, pbufUKM));
            // put the ukm in originator info then free it
            SME(AddUKMToOriginatorInfo(pbufUKM, &algTmp));
            delete (pbufUKM);
            // now load things into a new RecipientInfo
            if ((pRecipientInfo = m_SnaccEnvelopedData.recipientInfos.
                     Append()) == NULL)
               SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
            // load the encrypted key
            pRecipientInfo->encryptedKey.Set(bufEMEK.AccessAll(), 
                     bufEMEK.Length());
            // load the key encryption algorithm
            if ((pRecipientInfo->keyEncryptionAlgorithm = 
                     new KeyEncryptionAlgorithmIdentifier) == NULL)
               SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
            pRecipientInfo->keyEncryptionAlgorithm->algorithm.Set
                  (*((algTmp.GetId())->GetSNACCOid()));
            // store the km alg parameters
            if (bufParams.AccessAll() != NULL)
            {
               if ((pRecipientInfo->keyEncryptionAlgorithm->parameters = 
                        new AsnAny) == NULL)
                  SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
               SM_ASSIGN_ANYBUF((&bufParams), 
                        pRecipientInfo->keyEncryptionAlgorithm->parameters);
            }
            // load RecipientIdentifier rid
            if ((pRecipientInfo->rid = new RecipientIdentifier) == NULL)
               SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
            // TBD, hard code choice selection to issuer and serial number???
            pRecipientInfo->rid->choiceId == 
                     pRecipientInfo->rid->issuerAndSerialNumberCid;
            if ((pRecipientInfo->rid->issuerAndSerialNumber = 
                  new IssuerAndSerialNumber) == NULL)
               SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
            if ((pRecipientInfo->rid->issuerAndSerialNumber->issuer = 
                  new Name) == NULL)
               SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
            DECODE_BUF((pRecipientInfo->rid->issuerAndSerialNumber->issuer),
                  (*(pTmpRecipLst->Get()->GetIssuer()->GetEncodedDN())));
            //if ((pRecipientInfo->rid->issuerAndSerialNumber->issuer = new Name(
            //      *(Name*)(*(pTmpRecipLst->Get()->GetIssuer())))) == NULL)
            //   SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
            if ((pbufSerial = pTmpRecipLst->Get()->GetSerial()) == NULL)
               SME_THROW(SM_MEMORY_ERROR, NULL, NULL);
            pRecipientInfo->rid->issuerAndSerialNumber->serialNumber.
                  Set(pbufSerial->AccessAll(), pbufSerial->Length());

            // set the version of the recipient info
            // TBD, currently we set the version to zero because 
            // OriginatorCert is defaulted to absent and RecipientIdentifier
            // defaults to issuerAndSerialNumber
            pRecipientInfo->version = SM_ENV_DATA_PREV_VERSION;
            // TBD, eventually, add code to conditionally 
            // set SM_ENV_DATA_VERSION if conditions are met

            // now that we've created the new recipient info, we don't want to
            // continue searching through the instances so break out of the
            // instance loop and continue processing the recipients
            break;
         }
      } while ((pTmpInstLst = pTmpInstLst->Next()) != NULL);
   } while ((pTmpRecipLst = pTmpRecipLst->Next()) != NULL);
   
   // finished generating EMEKs, unlock the CTI
   SME(pCSMIME->InstanceUnlock(SM_INST_USE_THIS));

   ///////////////////////////////////////////////////////////
   // set the version for the enveloped data
   // loop through the recipient infos to find out if there is
   // a high version
   bool bFoundCurVersion = false;
   m_SnaccEnvelopedData.recipientInfos.SetCurrToFirst();
   pRecipientInfo = m_SnaccEnvelopedData.recipientInfos.First();
   while (pRecipientInfo != NULL)
   {
      if (pRecipientInfo->version == SM_ENV_DATA_VERSION)
         bFoundCurVersion == true;
      pRecipientInfo = m_SnaccEnvelopedData.recipientInfos.GoNext();
   }
   // from CMS: version in the syntax version number.  If originatorInfo
   // is present, then version shall be 2.  If any of the RecipientInfo
   // structures included have a version of 2, then the version shall be
   // 2.  If originatorInfo is absent and all of the RecipientInfo 
   // structures are version 0, then version shall be 0.
   if ((m_SnaccEnvelopedData.originatorInfo == NULL) &&
         (!bFoundCurVersion))
      m_SnaccEnvelopedData.version = SM_ENV_DATA_PREV_VERSION;
   else
      m_SnaccEnvelopedData.version = SM_ENV_DATA_VERSION;

   // finished filling m_SnaccEnvelopedData, ASN.1 encode it...
   ENCODE_BUF_NO_ALLOC((&m_SnaccEnvelopedData), pOutputBuf);

   // TBD, any local cleanup necessary?

   SME_FINISH
   SME_CATCH_SETUP
      // catch/cleanup logic as necessary
      // unlock the CTI
   SME_CATCH_FINISH
}

// EOF sm_Encrypt.cpp
